import { createQueryTester, $eq, $ne, $lt, $lte, $gt, $gte, $in, $nin, $all, $size, $regex, $elemMatch, $exists } from 'sift';

function wrapArray(value) {
  return Array.isArray(value) ? value : [value];
}
const TYPE_FIELD = '__caslSubjectType__';
function setSubjectType(type, object) {
  if (object) {
    if (!object.hasOwnProperty(TYPE_FIELD)) {
      Object.defineProperty(object, TYPE_FIELD, {
        value: type
      });
    } else if (type !== object[TYPE_FIELD]) {
      throw new Error(`Trying to cast object to subject type ${type} but previously it was casted to ${object[TYPE_FIELD]}`);
    }
  }

  return object;
}
function detectSubjectType(subject) {
  if (!subject) {
    return 'all';
  }

  if (typeof subject === 'string') {
    return subject;
  }

  if (subject.hasOwnProperty(TYPE_FIELD)) {
    return subject[TYPE_FIELD];
  }

  const Type = typeof subject === 'function' ? subject : subject.constructor;
  return Type.modelName || Type.name;
}
function expandActions(aliasMap, rawActions) {
  let actions = wrapArray(rawActions);
  let i = 0;

  while (i < actions.length) {
    const action = actions[i++];

    if (aliasMap.hasOwnProperty(action)) {
      actions = actions.concat(aliasMap[action]);
    }
  }

  return actions;
}

function assertAliasMap(aliasMap) {
  if (aliasMap.manage) {
    throw new Error('Cannot add alias for "manage" action because it is reserved');
  }

  Object.keys(aliasMap).forEach(alias => {
    const hasError = alias === aliasMap[alias] || Array.isArray(aliasMap[alias]) && (aliasMap[alias].indexOf(alias) !== -1 || aliasMap[alias].indexOf('manage') !== -1);

    if (hasError) {
      throw new Error(`Attempt to alias action to itself: ${alias} -> ${aliasMap[alias]}`);
    }
  });
}

const identity = x => x;
function createAliasResolver(aliasMap) {
  if (process.env.NODE_ENV !== 'production') {
    assertAliasMap(aliasMap);
  }

  return action => expandActions(aliasMap, action);
}

class Rule {
  constructor(rule, options) {
    this._matchConditions = void 0;
    this._matchField = void 0;
    this.action = options.resolveAction(rule.actions || rule.action);
    this.subject = rule.subject;
    this.inverted = !!rule.inverted;
    this.conditions = rule.conditions;
    this.reason = rule.reason;
    this.fields = !rule.fields || rule.fields.length === 0 ? undefined : wrapArray(rule.fields);

    if ('actions' in rule) {
      // eslint-disable-next-line
      console.warn('Rule `actions` field is deprecated. Use `action` field instead');
    }

    if (this.conditions && options.conditionsMatcher) {
      this._matchConditions = options.conditionsMatcher(this.conditions);
    }

    if (this.fields && options.fieldMatcher) {
      this._matchField = options.fieldMatcher(this.fields);
    }
  }

  matchesConditions(object) {
    if (!this._matchConditions) {
      return true;
    }

    if (!object || typeof object === 'string' || typeof object === 'function') {
      return !this.inverted;
    }

    return this._matchConditions(object);
  }

  matchesField(field) {
    if (!this._matchField) {
      return true;
    }

    if (!field) {
      return !this.inverted;
    }

    return this._matchField(field);
  }

}

class RulesAnalyzer {
  constructor(hasRules) {
    this._isAllInverted = void 0;
    this._hasPerFieldRules = false;
    this._hasRuleWithEmptyFields = false;
    this._isAllInverted = hasRules;
    this._analyze = this._analyze.bind(this);
  }

  _analyze({
    fields,
    inverted
  }) {
    this._isAllInverted = this._isAllInverted && !!inverted;

    if (!this._hasRuleWithEmptyFields && Array.isArray(fields) && !fields.length) {
      this._hasRuleWithEmptyFields = true;
    }

    if (!this._hasPerFieldRules && fields && fields.length) {
      this._hasPerFieldRules = true;
    }
  }

  _validate(ruleOptions) {
    if (this._isAllInverted) {
      // eslint-disable-next-line
      console.warn('Make sure your ability has direct rules, not only inverted ones. Otherwise `ability.can` will always return `false`.');
    }

    if (this._hasRuleWithEmptyFields) {
      // eslint-disable-next-line
      console.warn('[error in next major version]: There are rules with `fields` property being an empty array. This results in the same as not passing fields at all. Make sure to remove empty array or pass fields.');
    }

    if (this._hasPerFieldRules && !ruleOptions.fieldMatcher) {
      throw new Error('Field level restrictions are ignored because "fieldMatcher" option is not specified. Did you unintentionally used PureAbility instead of Ability?');
    }
  }

}

class RuleIndex {
  constructor(rules = [], options = {}) {
    this._hasPerFieldRules = false;
    this._mergedRules = Object.create(null);
    this._events = Object.create(null);
    this._indexedRules = Object.create(null);
    this._rules = [];
    this._ruleOptions = {
      conditionsMatcher: options.conditionsMatcher,
      fieldMatcher: options.fieldMatcher,
      resolveAction: options.resolveAction || identity
    };
    Object.defineProperty(this, 'detectSubjectType', {
      value: options.detectSubjectType || options.subjectName || detectSubjectType
    });
    Object.defineProperty(this, 'rules', {
      get: () => this._rules
    });
    this.update(rules);
  }

  update(rules) {
    const event = {
      rules,
      ability: this,
      target: this
    };

    this._emit('update', event);

    this._mergedRules = Object.create(null);
    const analyser = new RulesAnalyzer(rules.length > 0);

    const indexedRules = this._buildIndexFor(rules, analyser._analyze);

    analyser._validate(this._ruleOptions);

    this._indexedRules = indexedRules;
    this._rules = rules;
    this._hasPerFieldRules = analyser._hasPerFieldRules;

    this._emit('updated', event);

    return this;
  }

  _buildIndexFor(rawRules, iterator = identity) {
    const indexedRules = Object.create(null);

    for (let i = 0; i < rawRules.length; i++) {
      iterator(rawRules[i], i);
      const rule = new Rule(rawRules[i], this._ruleOptions);
      const priority = rawRules.length - i - 1;
      const actions = wrapArray(rule.action);
      const subjects = wrapArray(rule.subject);

      for (let k = 0; k < subjects.length; k++) {
        const subject = this.detectSubjectType(subjects[k]);
        indexedRules[subject] = indexedRules[subject] || Object.create(null);

        for (let j = 0; j < actions.length; j++) {
          const action = actions[j];
          indexedRules[subject][action] = indexedRules[subject][action] || Object.create(null);
          indexedRules[subject][action][priority] = rule;
        }
      }
    }

    return indexedRules;
  }

  possibleRulesFor(...args) {
    const [action, subject] = args;
    const subjectName = this.detectSubjectType(subject);
    const mergedRules = this._mergedRules;
    const key = `${subjectName}_${action}`;

    if (!mergedRules[key]) {
      mergedRules[key] = this._mergeRulesFor(action, subjectName);
    }

    return mergedRules[key];
  }

  _mergeRulesFor(action, subjectName) {
    const subjects = subjectName === 'all' ? [subjectName] : [subjectName, 'all'];
    const mergedRules = subjects.reduce((rules, subjectType) => {
      const subjectRules = this._indexedRules[subjectType];

      if (!subjectRules) {
        return rules;
      }

      return Object.assign(rules, subjectRules[action], subjectRules.manage);
    }, []); // TODO: think whether there is a better way to prioritize rules
    // or convert sparse array to regular one

    return mergedRules.filter(Boolean);
  }

  rulesFor(...args) {
    const [action, subject, field] = args;
    const rules = this.possibleRulesFor(action, subject);

    if (field && typeof field !== 'string') {
      throw new Error('The 3rd, `field` parameter is expected to be a string. See https://stalniy.github.io/casl/en/api/casl-ability#can-of-pure-ability for details');
    }

    if (!this._hasPerFieldRules) {
      return rules;
    }

    return rules.filter(rule => rule.matchesField(field));
  }

  on(event, handler) {
    const events = this._events;
    let isAttached = true;

    if (!events[event]) {
      events[event] = [];
    }

    events[event].push(handler);
    return () => {
      if (isAttached) {
        const index = events[event].indexOf(handler);
        events[event].splice(index, 1);
        isAttached = false;
      }
    };
  }

  _emit(eventName, event) {
    const handlers = this._events[eventName];

    if (handlers) {
      handlers.slice(0).forEach(handler => handler(event));
    }
  }

}

class PureAbility extends RuleIndex {
  can(...args) {
    const rule = this.relevantRuleFor(...args);
    return !!rule && !rule.inverted;
  }

  relevantRuleFor(...args) {
    const rules = this.rulesFor(...args);
    const subject = args[1];

    for (let i = 0; i < rules.length; i++) {
      if (rules[i].matchesConditions(subject)) {
        return rules[i];
      }
    }

    return null;
  }

  cannot(...args) {
    return !this.can(...args);
  }

}

const defaultOperations = {
  $eq,
  $ne,
  $lt,
  $lte,
  $gt,
  $gte,
  $in,
  $nin,
  $all,
  $size,
  $regex,
  $elemMatch,
  $exists
};
function buildMongoQueryMatcher(operations) {
  const options = {
    operations: Object.assign({}, defaultOperations, operations)
  };
  return conditions => createQueryTester(conditions, options);
}
const mongoQueryMatcher = buildMongoQueryMatcher({});

const REGEXP_SPECIAL_CHARS = /[-/\\^$+?.()|[\]{}]/g;
const REGEXP_ANY = /\.?\*+\.?/g;
const REGEXP_STARS = /\*+/;
const REGEXP_DOT = /\./g;

function detectRegexpPattern(match, index, string) {
  const quantifier = string[0] === '*' || match[0] === '.' && match[match.length - 1] === '.' ? '+' : '*';
  const matcher = match.indexOf('**') === -1 ? '[^.]' : '.';
  const pattern = match.replace(REGEXP_DOT, '\\$&').replace(REGEXP_STARS, matcher + quantifier);
  return index + match.length === string.length ? `(?:${pattern})?` : pattern;
}

function escapeRegexp(match, index, string) {
  if (match === '.' && (string[index - 1] === '*' || string[index + 1] === '*')) {
    return match;
  }

  return `\\${match}`;
}

function createPattern(fields) {
  const patterns = fields.map(field => field.replace(REGEXP_SPECIAL_CHARS, escapeRegexp).replace(REGEXP_ANY, detectRegexpPattern));
  const pattern = patterns.length > 1 ? `(?:${patterns.join('|')})` : patterns[0];
  return new RegExp(`^${pattern}$`);
}

const fieldPatternMatcher = fields => {
  let pattern;
  return field => {
    if (typeof pattern === 'undefined') {
      pattern = fields.every(f => f.indexOf('*') === -1) ? null : createPattern(fields);
    }

    return pattern === null ? fields.indexOf(field) !== -1 : pattern.test(field);
  };
};

class Ability extends PureAbility {
  constructor(rules, options) {
    super(rules, Object.assign({
      conditionsMatcher: mongoQueryMatcher,
      fieldMatcher: fieldPatternMatcher
    }, options));
  }

}

class RuleBuilder {
  constructor(rule) {
    this.rule = rule;
  }

  because(reason) {
    this.rule.reason = reason;
    return this;
  }

}

class AbilityBuilder {
  constructor(AbilityType = PureAbility) {
    this.rules = [];
    this._AbilityType = AbilityType;
    const self = this;
    self.can = self.can.bind(self);
    self.cannot = self.cannot.bind(self);
    self.build = self.build.bind(self);
  }

  can(action, subject, conditionsOrFields, conditions) {
    const rule = {
      action
    };

    if (subject) {
      rule.subject = subject;

      if (Array.isArray(conditionsOrFields) || typeof conditionsOrFields === 'string') {
        rule.fields = conditionsOrFields;
      } else if (typeof conditionsOrFields !== 'undefined') {
        rule.conditions = conditionsOrFields;
      }

      if (typeof conditions !== 'undefined') {
        rule.conditions = conditions;
      }
    }

    this.rules.push(rule);
    return new RuleBuilder(rule);
  }

  cannot(action, subject, conditionsOrFields, conditions) {
    const builder = this.can(action, subject, conditionsOrFields, conditions);
    builder.rule.inverted = true;
    return builder;
  }

  build(options) {
    return new this._AbilityType(this.rules, options);
  }

}
function defineAbility(params, dsl) {
  let options;
  let define;

  if (typeof params === 'function') {
    define = params;
    options = {};
  } else if (typeof dsl === 'function') {
    options = params;
    define = dsl;
  } else {
    throw new Error('`defineAbility` expects to receive either options and dsl function or only dsl function');
  }

  const builder = new AbilityBuilder(Ability);
  const result = define(builder.can, builder.cannot);
  return result && typeof result.then === 'function' ? result.then(() => builder.build(options)) : builder.build(options);
}

const getDefaultErrorMessage = error => `Cannot execute "${error.action}" on "${error.subjectType}"`;

const NativeError = function NError(message) {
  this.message = message;
};

NativeError.prototype = Object.create(Error.prototype);
class ForbiddenError extends NativeError {
  static setDefaultMessage(messageOrFn) {
    this._defaultErrorMessage = typeof messageOrFn === 'string' ? () => messageOrFn : messageOrFn;
  }

  static from(ability) {
    return new this(ability);
  }

  constructor(ability) {
    super('');
    this.ability = void 0;
    this.field = void 0;
    this.ability = ability;

    if (typeof Error.captureStackTrace === 'function') {
      this.name = 'ForbiddenError';
      Error.captureStackTrace(this, this.constructor);
    }
  }

  setMessage(message) {
    this.message = message;
    return this;
  }

  throwUnlessCan(...args) {
    const rule = this.ability.relevantRuleFor(...args);

    if (rule && !rule.inverted) {
      return;
    }

    this.action = args[0];
    this.subject = args[1];
    this.subjectType = this.ability.detectSubjectType(args[1]);
    this.field = args[2];
    const reason = rule ? rule.reason : ''; // eslint-disable-next-line no-underscore-dangle

    this.message = this.message || reason || this.constructor._defaultErrorMessage(this);
    throw this; // eslint-disable-line
  }

}
ForbiddenError._defaultErrorMessage = getDefaultErrorMessage;

export { Ability, AbilityBuilder, ForbiddenError, PureAbility, buildMongoQueryMatcher, createAliasResolver, defineAbility, detectSubjectType, fieldPatternMatcher, getDefaultErrorMessage, mongoQueryMatcher, setSubjectType as subject, wrapArray };
//# sourceMappingURL=index.js.map
